package main

import (
	"bytes"
	"encoding/json"
	"flag"
	"go/format"
	"io/ioutil"
	"log"
	"regexp"
	"sort"
	"strings"
	"text/template"
)

type options struct {
	jsonSource string
	template   string
	output     string
}

const version = "0.2.0"

func main() {
	log.SetFlags(log.Llongfile)
	opts := &options{
		template: errorsTemplate,
	}
	flag.StringVar(&opts.jsonSource, "json", "", `path to the ably-common/protocol/errors.json`)
	flag.StringVar(&opts.output, "o", "", "file to write output")
	flag.Parse()
	if opts.jsonSource == "" {
		flag.PrintDefaults()
		return
	}
	sourceFile, err := ioutil.ReadFile(opts.jsonSource)
	if err != nil {
		log.Fatal(err)
	}
	re := regexp.MustCompile(`\/\*([\s\S]*?)\*\/`)

	// we remove the multiline comments
	sourceFile = re.ReplaceAll(sourceFile, []byte{})
	codeToDesc := make(map[int]string)
	err = json.Unmarshal(sourceFile, &codeToDesc)
	if err != nil {
		log.Fatal(err)
	}
	var codes []int
	for k := range codeToDesc {
		codes = append(codes, k)
	}
	sort.Ints(codes)
	tpl, err := template.New("errors").Funcs(template.FuncMap{
		"key": func(k int) string {
			return codeToDesc[k]
		},
		"normalize": normalize,
		"split": func(a, b string) []string {
			return strings.Split(b, a)
		},
		"join": func(sep string, v []string) string {
			return strings.Join(v, sep)
		},
		"title": func(v []string) []string {
			for k := range v {
				v[k] = strings.Title(v[k])
			}
			return v
		},
		"map": func(from, to string, v []string) []string {
			dest := make([]string, len(v))
			for k := range v {
				if v[k] == from {
					dest = append(dest, to)
				} else {
					dest = append(dest, v[k])
				}
			}
			return dest
		},
	}).Parse(string(opts.template))
	if err != nil {
		log.Fatal(err)
	}
	var buf bytes.Buffer
	err = tpl.Execute(&buf, map[string]interface{}{
		"codes":   codes,
		"version": version,
	})
	if err != nil {
		log.Fatal(err)
	}
	goFile, err := format.Source(buf.Bytes())
	if err != nil {
		log.Fatal(err)
	}
	if opts.output != "" {
		err = ioutil.WriteFile(opts.output, goFile, 0600)
		if err != nil {
			log.Fatal(err)
		}
	}
}

func normalize(v string) string {
	idx := strings.Index(v, ")")
	if idx != -1 {
		v = v[:idx]
		v = strings.Replace(v, "(", " ", -1)
		v = strings.Replace(v, ")", " ", -1)
	}
	v = strings.Replace(v, "-", " ", -1)
	return v
}

const errorsTemplate = `// Code generated by errors-const v{{.version}} tool DO NOT EDIT.

package ably

// ErrorCode is the type for predefined Ably error codes.
type ErrorCode int

const(
	ErrNotSet ErrorCode = 0
	{{- range .codes}}
	{{template "constName" .}} ErrorCode = {{.}}
	{{- end}}
)

func (c ErrorCode) String() string {
	switch c {
	case ErrNotSet:
		return "(error code not set)"
	{{- range .codes}}
	case {{template "constName" .}}:
		return "{{key .}}"
	{{- end}}
	}
	return ""
}
{{- define "constName"}}Err{{key .|normalize|split " "|map "id" "ID"|map "clientId" "clientID"|map "tls" "TLS"|title|join ""}}{{- end}}
`
